'''
<2020.09.17>
Classify the model type to Tree(ensemble) based model and linear model.
'''

Tree = ['AdaBoost', 'Bagging', 'ExtraTrees', 'GradientBoosing', 'IsolationForest', 'RandomForest',
        'catboost', 'xgboost', 'LightGBM', 'HistGradientBoosting', 'DecisionTree', 'ExtraTree', 
        'XGBRegressor', 'XGBClassifier']
Linear = ['Linear', 'Logistic', 'Ridge', 'SGD', 'ElasticNet', 'Lars', 'Lasso',
          'OrthogonalMatchingPursuit', 'ARD', 'BayesianRidge', 'Huber', 'RANSAC', 'TheilSen',
          'Poisson', 'Tweedie', 'Gamma', 'PassiveAggresive', 'lars', 'lasso', 'orthogonal_mp',
          'ridge', 'LogisticRegression']




import inspect
import json
import warnings
import os
import re
from collections import OrderedDict
import gorilla
import sys
import logging
import boto3
from pathlib import Path
import numpy as np
from slack import WebClient
from slack.errors import SlackApiError
import keras
from keras.callbacks import CSVLogger
from Accuinsight.modeler.core import func, path, get
from Accuinsight.modeler.core.func import get_time
from Accuinsight.modeler.core.LcConst import LcConst
from Accuinsight.modeler.core.LcConst.LcConst import RUN_NAME_KERAS
from Accuinsight.modeler.core.get_for_visual import roc_pr_curve, get_visual_info_regressor
from Accuinsight.modeler.core.Run.RunInfo.RunInfo import set_current_runs, clear_runs, set_model_json_path, set_visual_csv_path, set_visual_json_path, set_best_model_h5_path, set_best_model_json_path, print_run_info, set_git_meta, set_python_dependencies, set_run_name, set_model_file_path, set_prefix_path
from Accuinsight.modeler.utils.dl_utils import delete_files_except_best, get_best_model_path
from Accuinsight.modeler.utils.runs_utils import get_aws_info, ProgressPercentage
from Accuinsight.modeler.utils.dependency.dependencies import gather_sources_and_dependencies
#from Accuinsight.modeler.core.notifier import notifierActor
from Accuinsight.modeler.utils.os_getenv import is_in_ipython, get_current_notebook
from Accuinsight.modeler.clients.modeler_api import LifecycleRestApi

logging.basicConfig(level=logging.INFO,
                    format='%(message)s')

warnings.filterwarnings("ignore")

class accuinsight(object):
    def __init__(self):
        self.BucketInfo = None
        self.endpoint = None
        self.thresholds = None
        self.token = None
        self.channel_id = None
        self.message = None

    def get_file(self, sub_path):
        self.BucketInfo = get_aws_info(sub_path)
        BUCKET_TYPE = self.BucketInfo['bucket_type']
        BUCKET_NAME = self.BucketInfo['bucket_name']
        FILE_PATH = self.BucketInfo['file_path']
        FILE_NAME = self.BucketInfo['file_name']
        FILE_TYPE = self.BucketInfo['file_type']
        FILE_DELIM = self.BucketInfo['file_delim']
        ACCESS_KEY = self.BucketInfo['my_access_key']
        SECRET_KEY = self.BucketInfo['my_secret_key']
        REGION = self.BucketInfo['region']
        URL = self.BucketInfo['endpoint']

        ## path for saving data
        save_dir = os.path.join(Path.home(), 'data_from_catalog')
        if os.path.exists(save_dir) == False:
            os.mkdir(save_dir)
        else:
            pass
        
        save_path = os.path.join(str(Path.home()), save_dir, FILE_NAME)
        
        # endpoint (수정)
        #pre_url = 'https://' + BUCKET_NAME + '.' + URL
        #self.endpoint = os.path.join(pre_url, FILE_PATH)

        ### data catalog 사용시
        # https://accuinsight-dev.cloudz.co.kr/datacatalog/#/management/table/detail/24/156
        # '/databases/24/tables/156'

        regex = re.compile('(\d+)/tables/(\d+)')
        host = 'https://accuinsight-dev.cloudz.co.kr/datacatalog/#/management/table/detail'
        self.endpoint = os.path.join(host, regex.search(sub_path).group(1), regex.search(sub_path).group(2))
        
        client = boto3.client(BUCKET_TYPE,
                              aws_access_key_id = ACCESS_KEY,
                              aws_secret_access_key = SECRET_KEY,
                              region_name = REGION)

        transfer = boto3.s3.transfer.S3Transfer(client)
               
        progress = ProgressPercentage(client, BUCKET_NAME, FILE_PATH)

        sys.stdout.write('%s %s %s' % ('Downloading file...', FILE_NAME,  '\n'))
        transfer.download_file(BUCKET_NAME, FILE_PATH, save_path, callback=progress)
        logging.info(save_path)
        
    def set_slack(self, token = None, channel_id = None):
        self.token = token
        self.channel_id = channel_id

    def send_message(self, message = None, thresholds = None):
        if message is not None and thresholds is not None:
            raise ValueError("'message'와 'thresholds' 두 개의 arguments를 동시에 입력할 수 없습니다.")
        else:
            self.message = message
            self.thresholds = thresholds

    def autolog(self, tag = None, best_weights = False):
        global description, endpoint, var_model_file_path, message, thresholds, token, channel_id
        description = tag
        endpoint = self.endpoint
        message = self.message
        thresholds = self.thresholds
        token = self.token
        channel_id = self.channel_id
        if best_weights == True:
            best_weights_on = True
        else:
            best_weights_on = False

        
        is_notebook = is_in_ipython()
        if is_notebook == True:
            var_model_file_path = get_current_notebook()
            _caller_globals = inspect.stack()[1][0].f_globals
            (
                mainfile,
                sources,
                dependencies
            ) = gather_sources_and_dependencies(
                globs=_caller_globals,
                save_git_info=False
            )
        else:
            _caller_globals = inspect.stack()[1][0].f_globals
            (
                mainfile,
                sources,
                dependencies
            ) = gather_sources_and_dependencies(
                globs=_caller_globals,
                save_git_info=True
            )
            var_model_file_path = mainfile['filename']

    
        class TrainHistoryCallbacks(keras.callbacks.Callback):
            def __init__(self, verbose = 1, mode = 'auto', period = 1):
                super(TrainHistoryCallbacks, self).__init__()
                self.verbose = verbose
                self.period = period
                self.best_epochs = 0
                self.epochs_since_last_save = 0
                self.mode = mode
        
            def on_train_begin(self, logs={}):
                if best_weights_on:
                    logging.info('Using autolog(best_weights=True)')
                else:
                    logging.info('Using autolog(best_weights=False)')
            
                global start
                start = get_time.now()
                self.run_id = func.get_run_id()
                self.model_type = get.model_type(self.model)
                self.model_summary = OrderedDict()
                self.model_summary['data_version'] = endpoint
                self.model_summary['model_description'] = description
                self.model_summary['logging_time'] = get_time.logging_time()
                self.model_summary['run_id'] = func.get_run_id()
                self.model_summary['model_type'] = get.model_type(self.model)
                
                if hasattr(self.model.loss, 'get_config'):
                    self.model_summary['loss_function'] = self.model.loss.get_config()['name']
                else:
                    self.model_summary['loss_function'] = self.model.loss
                
                self.model_summary['optimizer_info'] = {type(self.model.optimizer).__name__: self.model.optimizer.get_config()}
            
            
                '''[get best model] on_train_begin '''
                self.best_weights = self.model.get_weights()
              
                self.dict_path = path.get_file_path(self.model, usedFramework='keras')
            
                set_prefix_path(self.dict_path[LcConst.RUN_PREFIX_PATH])
            
                set_run_name(self.model_type, self.run_id)
                set_python_dependencies(py_depenpency=dependencies)
            
            '''[get best model] on_epoch_end '''
            def on_epoch_end(self, epoch, logs=None):
                logs = logs or {}
                
                if epoch == 0:
                    if len(self.model.metrics_names) == 1 and 'loss' in self.model.metrics_names:
                        self.monitor = 'val_loss'
                    elif len(self.model.metrics_names) >= 2:
                        self.monitor = 'val_' + self.model.metrics_names[1]
                    # set monitoring option
                    if self.mode not in ['auto', 'min', 'max']:
                        warnings.warn('GetBest mode %s is unknown, '
                                      'fallback to auto mode.' % (self.mode), RuntimeWarning)
                        self.mode = 'auto'
                    if self.mode == 'min':
                        self.monitor_op = np.less
                        self.best = np.Inf
                    elif self.mode == 'max':
                        self.monitor_op = np.greater
                        self.best = -np.Inf
                    else:
                        if 'acc' in self.monitor or 'f1' in self.monitor:
                            self.monitor_op = np.greater
                            self.best = -np.Inf
                        else:
                            self.monitor_op = np.less
                            self.best = np.Inf
                else:
                    pass
                
                # Using best_weights
                if best_weights_on:
                    
                    # update best_weights
                    self.epochs_since_last_save += 1
                    if self.epochs_since_last_save >= 1:
                        self.epochs_since_last_save = 0
                        current = logs.get(self.monitor)
                        if current is None:
                            warnings.warn('Can pick best model only with %s available, '
                                      'skipping.' % (self.monitor), RuntimeWarning)
                        else:
                            if self.monitor_op(current, self.best):
#                                 if self.verbose > 0:
#                                     print('\n\nEpoch %05d: %s improved from %0.5f to %0.5f,'
#                                           ' storing weights.\n'
#                                           % (epoch + 1, self.monitor, self.best, current))
                                self.best = current
                                self.best_epochs = epoch + 1
                                self.best_weights = self.model.get_weights()
                            else:
                                pass
#                                 if self.verbose > 0:
#                                     print('\n\nEpoch %05d: %s did not improve\n' % (epoch + 1, self.monitor))

                    self.current_value = current
                    
                    # model save path
                    common_path = self.dict_path['save_model_path'] + self.model_summary['model_type'] + '-' + \
                    self.model_summary['run_id'] + '-epoch-' + str(epoch + 1).zfill(5) + '-' + self.monitor + '-' + str(current).zfill(5)
                    save_model_path = common_path + '.json'
                    save_weights_path = common_path + '.h5'

                    # model to JSON
                    model_json = self.model.to_json()
                    with open(save_model_path, "w") as json_file:
                        json_file.write(model_json)
                    # weights to H5
                    self.model.save_weights(save_weights_path)
    
                # Not using best_weights
                else:
                    self.last_epochs = epoch + 1
                    self.last_epoch_metric = logs.get(self.monitor)
                    
                    # model save path
                    common_path = self.dict_path['save_model_path'] + self.model_summary['model_type'] + '-' + \
                    self.model_summary['run_id'] + '-epoch-' + str(epoch + 1).zfill(5) + '-' + self.monitor + '-' + str(self.last_epoch_metric).zfill(5)
                    save_model_path = common_path + '.json'
                    save_weights_path = common_path + '.h5'

                    # model to JSON
                    model_json = self.model.to_json()
                    with open(save_model_path, "w") as json_file:
                        json_file.write(model_json)
                    # weights to H5
                    self.model.save_weights(save_weights_path)
                    
                    self.current_value = self.last_epoch_metric
            
            def on_train_end(self, logs={}):
                '''[get best model] on_train_end '''
                if self.verbose > 0:
                    print('Using epoch %05d with %s: %0.5f' % (self.best_epochs, self.monitor, self.best))
                self.model.set_weights(self.best_weights)  # set best model's weights
                
#                self.model_summary = OrderedDict()
#                self.model_summary['data_version'] = endpoint
#                self.model_summary['model_description'] = description
#                self.model_summary['logging_time'] = get_time.logging_time()
#                self.model_summary['run_id'] = self.run_id
#                self.model_summary['model_type'] = self.model_type
#
#                if hasattr(self.model.loss, 'get_config'):
#                    self.model_summary['loss_function'] = self.model.loss.get_config()['name']
#                else:
#                    self.model_summary['loss_function'] = self.model.loss
#
#                self.model_summary['optimizer_info'] = {type(self.model.optimizer).__name__: self.model.optimizer.get_config()}
                
                end = get_time.now()
                self.model_summary['time_delta'] = str(end - start)
                self.model_summary['selected_metrics'] = {self.monitor: self.best}
            
                # path for model_info.json
                self.path_for_setting_model_json = self.dict_path['model_json']
                set_model_json_path(self.path_for_setting_model_json)
            
                model_json_full_path = self.dict_path['model_json_full']
            
                with open(model_json_full_path, 'w', encoding='utf-8') as save_file:
                    json.dump(self.model_summary, save_file, indent="\t")
            
                if best_weights_on:
                    delete_files_except_best(run_id=self.model_summary['run_id'], epochs=str(self.best_epochs),
                                            path=self.dict_path)
                else:
                    delete_files_except_best(run_id=self.model_summary['run_id'], epochs=str(self.last_epochs),
                    path=self.dict_path)

                path_for_setting_model_json = self.dict_path['save_model_dir'] + get_best_model_path(run_id = self.run_id, path = self.dict_path)['json']
                set_best_model_json_path(path_for_setting_model_json)
                path_for_setting_model_h5 = self.dict_path['save_model_dir'] + get_best_model_path(run_id = self.run_id, path = self.dict_path)['h5']
                set_best_model_h5_path(path_for_setting_model_h5)
            
            
                start_ts = int(start.timestamp())
                end_ts = int(end.timestamp())
                delta_ts = end_ts - start_ts
            
                clear_runs(start_ts, end_ts, delta_ts)
                accuinsight._send_message(metric = self.monitor,
                                          current_value = self.current_value,
                                          message = message,
                                          thresholds = thresholds,
                                          token = token,
                                          channel_id = channel_id)
                modeler_rest = LifecycleRestApi(LcConst.BACK_END_API_URL,
                                                LcConst.BACK_END_API_PORT,
                                                LcConst.BACK_END_API_URI)
                modeler_rest.lc_create_run()
                accuinsight.off_autolog()
            
        
        class visualCallbacks(keras.callbacks.Callback):
            def __init__(self, x_validation=None, y_validation=None):
                self.x_val = x_validation
                self.y_val = y_validation

            def on_train_end(self, logs={}):

                self.dict_path = path.get_file_path(self.model, usedFramework='keras')

                # path for visual.json
                path_for_setting_visual_json = self.dict_path['visual_json']
                visual_json_full_path = self.dict_path['visual_json_full']
                set_visual_json_path(path_for_setting_visual_json)

                # classification
                if get.is_classification(self.model):
                    visual_classification_json = roc_pr_curve(self.model, self.x_val, self.y_val)

                    with open(visual_json_full_path, 'w', encoding='utf-8') as save_file:
                        json.dump(visual_classification_json, save_file, indent="\t")

                # regression
                else:
                    visual_regression_json = OrderedDict()
                    visual_regression_json['True_y'] = self.y_val.tolist()
                    visual_regression_json['Predicted_y'] = get_visual_info_regressor(self.model, self.x_val)

                    with open(visual_json_full_path, 'w', encoding='utf-8') as save_file:
                        json.dump(visual_regression_json, save_file, indent="\t")
            
        def run_and_log_function(self, original, args, kwargs, unlogged_params):

            dict_path = path.get_file_path(self, usedFramework='keras')
        
            path_for_setting_visual_csv = dict_path['visual_csv']
            visual_csv_full_path = dict_path['visual_csv_full']
        
            # set current run
            set_current_runs(RUN_NAME_KERAS)
            set_model_file_path(var_model_file_path)

            set_visual_csv_path(path_for_setting_visual_csv)

            csv_logger = CSVLogger(visual_csv_full_path, append=True, separator=';')
 
 
            ''' save json for visualization '''
            kwargs_dict = OrderedDict()
            
            for key, value in kwargs.items():
                kwargs_dict[key] = value
            
            # using validation_data argument
            if 'validation_data' in kwargs_dict.keys():
                validation_set = kwargs['validation_data']

                try:
                    x_val = validation_set[0]
                    y_val = validation_set[1]

                except:
                    iterator = iter(validation_set)
                    valid_set = next(iterator)
                    x_val = valid_set[0].numpy()
                    y_val = valid_set[1].numpy()
            
#            elif 'validation_split' in kwargs_dict.keys():
#                (x, y), validation_set = (data_adapter.train_validation_split((x,y), validation_split=kwargs_dict['validation_split']))
#                if validation_set:
#                    x_val, y_val, val_sample_weight = (data_adapter.unpack_x_y_sample_weight(validation_set))
#
            else:
                raise ValueError('"validation_data" or "validation_split" does not exist.')
            
            get_visual = visualCallbacks(x_validation=x_val, y_validation=y_val)


            if 'callbacks' in kwargs:
                kwargs['callbacks'] += [csv_logger]
                kwargs['callbacks'] += [get_visual]
                kwargs['callbacks'] += [TrainHistoryCallbacks()]
            
            else:
                kwargs['callbacks'] = [csv_logger]
                kwargs['callbacks'] += [get_visual]
                kwargs['callbacks'] += [TrainHistoryCallbacks()]

            fit_model = original(self, *args, **kwargs)
            return(fit_model)

        @gorilla.patch(keras.Model)
        def fit(self, *args, **kwargs):
            original = gorilla.get_original_attribute(keras.Model, 'fit')
            unlogged_params = ['self', 'x', 'y', 'callbacks', 'validation_data', 'verbose']
            return run_and_log_function(self, original, args, kwargs, unlogged_params)

        settings = gorilla.Settings(allow_hit=True, store_hit=True)
        gorilla.apply(gorilla.Patch(keras.Model, 'fit', fit, settings=settings))


    def _send_message(metric = None, current_value = None, message = None, thresholds=None, token = None, channel_id = None):
        if token is not None and channel_id is not None:
            if thresholds is not None:
                try:
                    if current_value >= thresholds:
                        msg = '[모델 학습 완료] ' + metric +'이 설정하신 thresolds: ' + str(thresholds) + '를 초과하였습니다.'
                        response = WebClient(token=token).chat_postMessage(channel=channel_id, text=msg)
                except SlackApiError as e:
                    assert e.response["error"]
            elif message is not None:
                try:
                    msg = message
                    response = WebClient(token=token).chat_postMessage(channel=channel_id, text=msg)
                except SlackApiError as e:
                    assert e.response["error"]
            else:
                pass

    def off_autolog():
        def stop_log(self, original, args, kwargs, unlogged_params):
            fit_model = original(self, *args, **kwargs)
            return(fit_model)
        
        @gorilla.patch(keras.Model)
        def fit(self, *args, **kwargs):
            original = gorilla.get_original_attribute(keras.Model, 'fit')
            unlogged_params = ['self', 'x', 'y', 'callbacks', 'validation_data', 'verbose']
            return stop_log(self, original, args, kwargs, unlogged_params)

        settings = gorilla.Settings(allow_hit=True, store_hit=True)
        gorilla.apply(gorilla.Patch(keras.Model, 'fit', fit, settings=settings))



def experimental(func):
    """
    Decorator for marking APIs experimental in the docstring.

    :param func: A function to mark
    :returns Decorated function.
    """
    notice = ".. Note:: Experimental: This method may change or " + \
             "be removed in a future release without warning.\n"
    func.__doc__ = notice + func.__doc__
    return func




import warnings
from collections import OrderedDict
import numpy as np
from sklearn.preprocessing import OneHotEncoder
from sklearn.metrics import confusion_matrix, f1_score, precision_score, recall_score
from Accuinsight.modeler.core import get
from Accuinsight.modeler.core.func_for_visual import cal_confusion, cal_auc, multi_roc, multi_pr


def y_test_reshape(y_test):
    # dataframe/series형 np.array로 형 변환
    if 'pandas' in str(type(y_test)):
        y_test = np.array(y_test)
        
    try:
        n_classes = len(np.unique(y_test, axis = 0))
    except:
        n_classes = len(np.unique(y_test))

    if y_test.ndim == 1:
        y_test = y_test
    elif y_test.ndim == 2:
        if y_test.shape[1] == n_classes:
            y_test = np.array([np.where(y_test[i]==np.max(y_test[i]))[0].tolist() for i in range(len(y_test))]).flatten()
        else:
            y_test = y_test.flatten()
    return(y_test)

def y_test_to_categorical(y_test):
    try:
        n_classes = len(np.unique(y_test, axis = 0))
    except:
        n_classes = len(np.unique(y_test))
        
    enc = OneHotEncoder(categories = 'auto')

    if y_test.ndim == 1:
        y_test2d = y_test.reshape(-1,1)
        y_test_cat = enc.fit_transform(y_test2d).toarray()
    
    elif y_test.ndim == 2 and y_test.shape[1] == n_classes:  # y_test: categorical
        y_test_cat = y_test

    return(y_test_cat)

def y_pred_reshape(model, X_test):
    prob = model.predict(X_test)

    if prob.ndim == 1:
        y_pred = prob

    elif prob.ndim == 2:
        if prob.shape[1] == 1:
            y_prob = np.c_[1-prob, prob]
            y_pred = np.array([np.where(y_prob[i] == np.max(y_prob[i]))[0].tolist() for i in range(len(y_prob))]).flatten()
        else:
            y_prob = prob
            y_pred = np.array([np.where(y_prob[i]==np.max(y_prob[i]))[0].tolist() for i in range(len(y_prob))]).flatten()
    return(y_pred)

def y_pred_to_categorical(model, X_test, y_test):
    try:
        n_classes = len(np.unique(y_test, axis = 0))
    except:
        n_classes = len(np.unique(y_test))
        
    prob = model.predict(X_test)
    enc = OneHotEncoder(categories = 'auto')

    if prob.ndim == 1:
        y_pred2d = prob.reshape(-1,1)
        y_pred_cat = enc.fit_transform(y_pred2d).toarray()

    elif prob.ndim == 2:
        if n_classes == 2 and prob.shape[1] == n_classes-1:
            prob = np.c_[1-prob, prob]
            y_pred_cat = np.zeros_like(prob)
            for i in range(y_pred_cat.shape[0]):
                y_pred_cat[i][np.where(prob[i] == np.max(prob[i]))[0]] = 1
    
        elif prob.shape[1] == n_classes:  # y_test: categorical
            y_pred_cat = np.zeros_like(prob)
            for i in range(y_pred_cat.shape[0]):
                y_pred_cat[i][np.where(prob[i] == np.max(prob[i]))[0]] = 1

    return(y_pred_cat)

def y_score_proba(model, X_test, y_test):
    if get.model_type(model) == 'SVC' and 'probability=False' in str(model):
        if len(model.classes_) == 2:
            d = model.decision_function(X_test)
            prob = (d - d.min()) / (d.max() - d.min())
            y_score = np.c_[prob, 1-prob]
        else:
            warnings.warn('"probability = True" 설정이 필요합니다.')

    elif get.model_type(model) == 'LinearSVC':
        if len(model.classes_) == 2:
            d = model.decision_function(X_test)
            prob = (d - d.min()) / (d.max() - d.min())
            y_score = np.c_[prob, 1-prob]
        else:
            warnings.warn('label의 갯수가 3 이상인 경우, LinearSVC에서 "predict_proba"를 제공하지 않아 metric 계산이 불가능합니다. 자세한 사항은 사용자 매뉴얼을 참조하십시오.')
            y_score = None
            
    elif 'keras' in get.model_type(model):
#       prob = model.predict_proba(X_test)   -- predict_proba() is deprecated.
        prob = model.predict(X_test)
        if len(np.unique(y_test, axis = 0)) == 2 and prob.shape[1] == 1:
            y_score = np.c_[1-prob, prob]
            
        else:
            y_score = prob

    else:
        y_score = model.predict_proba(X_test)
    return(y_score)

# (DL & ML) & Classification
def roc_pr_curve(model, X_test, y_test, *args):
    
    try:
        n_classes = len(np.unique(y_test, axis = 0))
    except:
        n_classes = len(np.unique(y_test))
    
    # y_test (dim=2/dim=1, 두 가지 경우 존재)
    # y_test는 categorical var. 변경 필요
    y_test = y_test_reshape(y_test)
    y_test_cat = y_test_to_categorical(y_test)

    # y_score (SVC 사용한 경우, decision_function으로 계산)
    y_score = y_score_proba(model, X_test, y_test)

    # y_pred (dim=2/dim=1, 두 가지 경우 존재)
    y_pred = y_pred_reshape(model, X_test)

    # list of labels
    if hasattr(model, 'classes_'):   # case 1: ML
        if 'float' in str(model.classes_.dtype) or 'int' in str(model.classes_.dtype):
            label_list = np.unique(model.classes_).astype('int').tolist()
        else:
            label_list = np.unique(model.classes_).tolist()
    else:   # case 2: DL
        classes_dl = np.unique(y_test, axis = 0)
        if 'float' in str(classes_dl.dtype) or 'int' in str(classes_dl.dtype):
            label_list = classes_dl.astype('int').tolist()
        else:
            label_list = classes_dl.tolist()

    # results
    res_dict = OrderedDict()
    
    if y_score is not None:
        res_roc = multi_roc(y_test_cat.T, y_score.T)
        res_pr = multi_pr(y_test_cat.T, y_score.T)
    else:
        res_dict['ValueError'] = True

    res_dict['fpr'] = res_roc['fpr']
    res_dict['tpr'] = res_roc['tpr']
    res_dict['recall'] = res_pr['recall']
    res_dict['precision'] = res_pr['precision']

    res_dict['legend'] = {'roc': res_roc['auc'], 'precision-recall': res_pr['auc']}
    res_dict['confusion_matrix'] = dict(zip(label_list, confusion_matrix(y_test, y_pred, labels = label_list).tolist()))
    res_dict['chart'] = {'AUC': res_roc['auc']['micro'],
                         'Precision': res_pr['auc']['micro'],
                         'Recall': recall_score(y_test, y_pred, average = 'micro'),
                         'F1-score': f1_score(y_test, y_pred, average = 'micro')}
    return(res_dict)

# (DL & ML) & Regression
def get_true_y(y_test):
    
    # reshape (dim2 -> dim1)
    if len(y_test.shape) == 2:
        true_y = y_test.reshape(1,-1)[0].tolist()
    elif type(y_test) == list:
        true_y = y_test
    else:
        true_y = y_test.tolist()
    return(true_y)
        
def get_visual_info_regressor(model, X_test):
    preds = model.predict(X_test)

    # reshape (dim2 -> dim1)
    if len(preds.shape) == 2:
        preds = preds.reshape(1,-1)[0]

    # 소수점 자리수 조정(소수점 이하 5번째 자리에서 반올림)
    try:
        if len(str(preds[0]).split('.')[1]) >= 5:
            preds = np.round(preds, 4)
            preds = list(map(float, np.array(preds, dtype = np.str).tolist()))
    except TypeError:
        pass
    return(preds)
